@@ -8,7 +8,7 @@
  *  The iODBC driver manager.
  *
  *  Copyright (C) 1995 Ke Jin <kejin@empress.com>
- *  Copyright (C) 1996-2023 OpenLink Software <iodbc@openlinksw.com>
+ *  Copyright (C) 1996-2021 OpenLink Software <iodbc@openlinksw.com>
  *  All Rights Reserved.
  *
  *  This software is released under the terms of either of the following
@@ -85,7 +85,7 @@
 #include <stdlib.h>
 #include <unistd.h>
 
-#include "unicode.h"
+#include <unicode.h>
 
 #include <fcntl.h>
 #include <sys/stat.h>
