@@ -7,7 +7,7 @@
  *
  *  The iODBC driver manager.
  *
- *  Copyright (C) 1996-2014 by OpenLink Software <iodbc@openlinksw.com>
+ *  Copyright (C) 1996-2021 OpenLink Software <iodbc@openlinksw.com>
  *  All Rights Reserved.
  *
  *  This software is released under the terms of either of the following
@@ -114,6 +114,88 @@ SQLRETURN SQLAllocStmt_Internal (SQLHDBC hdbc, SQLHSTMT * phstmt);
 SQLRETURN SQLFreeStmt_Internal (SQLHSTMT hstmt, SQLUSMALLINT fOption);
 SQLRETURN SQLTransact_Internal (SQLHENV henv, SQLHDBC hdbc, SQLUSMALLINT fType);
 
+SQLRETURN
+SQLAllocDesc_Internal (
+    SQLHDBC hdbc,
+    SQLHDESC * phdesc)
+{
+  CONN (pdbc, hdbc);
+  SQLRETURN retcode = SQL_SUCCESS;
+  HPROC hproc = SQL_NULL_HPROC;
+  DESC_t *new_desc;
+  SQLUINTEGER odbc_ver;
+  SQLUINTEGER dodbc_ver;
+
+  /* check state */
+  switch (pdbc->state)
+    {
+    case en_dbc_connected:
+    case en_dbc_hstmt:
+      break;
+
+    case en_dbc_allocated:
+    case en_dbc_needdata:
+      PUSHSQLERR (pdbc->herr, en_08003);
+      *phdesc = SQL_NULL_HSTMT;
+
+      return SQL_ERROR;
+
+    default:
+      return SQL_INVALID_HANDLE;
+    }
+
+  odbc_ver = ((GENV_t *) pdbc->genv)->odbc_ver;
+  dodbc_ver = (pdbc->henv != SQL_NULL_HENV) ? ((ENV_t *) pdbc->henv)->dodbc_ver : odbc_ver;
+
+  if (odbc_ver == SQL_OV_ODBC2 || dodbc_ver == SQL_OV_ODBC2)
+    {
+      PUSHSQLERR (pdbc->herr, en_HYC00);
+      return SQL_ERROR;
+    }
+
+  if (phdesc == NULL)
+    {
+      PUSHSQLERR (pdbc->herr, en_HY009);
+      return SQL_ERROR;
+    }
+
+  hproc = _iodbcdm_getproc (pdbc, en_AllocHandle);
+
+  if (hproc == SQL_NULL_HPROC)
+    {
+      PUSHSQLERR (pdbc->herr, en_IM001);
+      return SQL_ERROR;
+    }
+
+  new_desc = (DESC_t *) MEM_ALLOC (sizeof (DESC_t));
+  if (!new_desc)
+    {
+      PUSHSQLERR (pdbc->herr, en_HY001);
+      return SQL_ERROR;
+    }
+  memset (new_desc, 0, sizeof (DESC_t));
+  CALL_DRIVER (pdbc, pdbc, retcode, hproc, (SQL_HANDLE_DESC, pdbc->dhdbc, &new_desc->dhdesc));
+
+  if (!SQL_SUCCEEDED (retcode))
+    {
+      MEM_FREE (new_desc);
+      return SQL_ERROR;
+    }
+
+  new_desc->type = SQL_HANDLE_DESC;
+  new_desc->hdbc = pdbc;
+  new_desc->hstmt = NULL;
+  new_desc->herr = NULL;
+  new_desc->desc_cip = 0;
+  *phdesc = new_desc;
+
+  new_desc->next = (DESC_t *) pdbc->hdesc;
+  pdbc->hdesc = new_desc;
+
+  return SQL_SUCCESS;
+}
+
+
 RETCODE SQL_API
 SQLAllocHandle_Internal (
     SQLSMALLINT	  handleType,
@@ -159,9 +241,6 @@ SQLAllocHandle_Internal (
     case SQL_HANDLE_DESC:
       {
 	CONN (con, inputHandle);
-	HPROC hproc = SQL_NULL_HPROC;
-	RETCODE retcode;
-	DESC_t *new_desc;
 
 	if (!IS_VALID_HDBC (con))
 	  {
@@ -169,52 +248,7 @@ SQLAllocHandle_Internal (
 	  }
 	CLEAR_ERRORS (con);
 
-	if (((ENV_t *)(con->henv))->dodbc_ver == SQL_OV_ODBC2)
-	  {
-	    PUSHSQLERR (con->herr, en_HYC00);
-	    return SQL_ERROR;
-	  }
-	if (!outputHandlePtr)
-	  {
-	    PUSHSQLERR (con->herr, en_HY009);
-	    return SQL_ERROR;
-	  }
-
-	hproc = _iodbcdm_getproc (con, en_AllocHandle);
-
-	if (hproc == SQL_NULL_HPROC)
-	  {
-	    PUSHSQLERR (con->herr, en_IM001);
-	    return SQL_ERROR;
-	  }
-
-	new_desc = (DESC_t *) MEM_ALLOC (sizeof (DESC_t));
-	if (!new_desc)
-	  {
-	    PUSHSQLERR (con->herr, en_HY001);
-	    return SQL_ERROR;
-	  }
-	memset (new_desc, 0, sizeof (DESC_t));
-	CALL_DRIVER (con, con, retcode, hproc,
-	    (handleType, con->dhdbc, &new_desc->dhdesc));
-
-	if (!SQL_SUCCEEDED (retcode))
-	  {
-	    MEM_FREE (new_desc);
-	    return SQL_ERROR;
-	  }
-
-	new_desc->type = SQL_HANDLE_DESC;
-	new_desc->hdbc = con;
-	new_desc->hstmt = NULL;
-	new_desc->herr = NULL;
-	new_desc->desc_cip = 0;
-	*outputHandlePtr = new_desc;
-
-	new_desc->next = (DESC_t *) con->hdesc;
-	con->hdesc = new_desc;
-
-	return SQL_SUCCESS;
+        return SQLAllocDesc_Internal (inputHandle, outputHandlePtr);
       }
 
     default:
@@ -586,6 +620,28 @@ SQLSetEnvAttr_Internal (SQLHENV environmentHandle,
 	  return SQL_ERROR;
 	}
 
+    case SQL_ATTR_APP_UNICODE_TYPE:
+      switch ((SQLINTEGER)(SQLULEN) ValuePtr)
+        {
+        case SQL_DM_CP_UTF16:
+          genv->conv.dm_cp = CP_UTF16;
+          break;
+        case SQL_DM_CP_UCS4:
+          genv->conv.dm_cp = CP_UCS4;
+          break;
+        case SQL_DM_CP_UTF8:
+          genv->conv.dm_cp = CP_UTF8;
+          break;
+        default:
+	  PUSHSQLERR (genv->herr, en_HY024);
+	  return SQL_ERROR;
+        }
+      DPRINTF ((stderr,
+      "DEBUG: SQLSetEnvAttr DiverManager AppUnicodeType=%s\n",
+        genv->conv.dm_cp==CP_UCS4?"UCS4":(genv->conv.dm_cp==CP_UTF16?"UTF16":"UTF8")));
+      
+      return SQL_SUCCESS;
+
     default:
       PUSHSQLERR (genv->herr, en_HY092);
       return SQL_ERROR;
@@ -642,6 +698,7 @@ SQLSetEnvAttr (
 	    EnvironmentHandle,
 	    Attribute,
 	    ValuePtr, StringLength));
+      return retcode;
    }
 }
 
@@ -1349,7 +1406,11 @@ SQLSetStmtAttr_Internal (
           return SQL_ERROR;
         }
       if (Attribute == SQL_ATTR_PARAM_BIND_TYPE)
-        stmt->bind_type = (SQLUINTEGER) (SQLULEN) ValuePtr;
+        stmt->param_bind_type = (SQLUINTEGER) (SQLULEN) ValuePtr;
+      else if (Attribute == SQL_ATTR_PARAM_BIND_OFFSET_PTR)
+        stmt->param_bind_offset = (SQLUINTEGER) (SQLULEN) ValuePtr;
+      else if (Attribute == SQL_ATTR_ROW_BIND_OFFSET_PTR)
+        stmt->row_bind_offset = (SQLUINTEGER) (SQLULEN) ValuePtr;
       return retcode;
 
     case SQL_ATTR_ROWS_FETCHED_PTR:
@@ -1452,7 +1513,7 @@ SQLSetStmtAttr_Internal (
 
           if (SQL_SUCCEEDED (retcode))
             {
-              stmt->rowset_size = Attribute;
+              stmt->rowset_size = (SQLUINTEGER) (SQLULEN) ValuePtr;
               if (retcode == SQL_SUCCESS_WITH_INFO)
                 {
                   SQLUINTEGER data;
@@ -1497,7 +1558,7 @@ SQLSetStmtAttr_Internal (
 
               if (SQL_SUCCEEDED (retcode))
                 {
-                  stmt->rowset_size = Attribute;;
+                  stmt->rowset_size = (SQLUINTEGER) (SQLULEN) ValuePtr;
                   if (retcode == SQL_SUCCESS_WITH_INFO)
                     {
                       SQLUINTEGER data;
@@ -1517,7 +1578,7 @@ SQLSetStmtAttr_Internal (
 
               if (SQL_SUCCEEDED (retcode))
                 {
-                  stmt->rowset_size = Attribute;;
+                  stmt->rowset_size = (SQLUINTEGER) (SQLULEN) ValuePtr;
                   if (retcode == SQL_SUCCESS_WITH_INFO)
                     {
                       SQLUINTEGER data;
@@ -1591,7 +1652,7 @@ SQLSetStmtAttr_Internal (
             }
 
           if (Attribute == SQL_ATTR_ROW_BIND_TYPE && SQL_SUCCEEDED (retcode))
-            stmt->bind_type = Attribute;
+            stmt->row_bind_type = (SQLUINTEGER) (SQLULEN) ValuePtr;
 
           return retcode;
 	}
@@ -1603,7 +1664,7 @@ SQLSetStmtAttr_Internal (
 	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
 		  (stmt->dhstmt, Attribute, ValuePtr));
               if (Attribute == SQL_ATTR_ROW_BIND_TYPE && SQL_SUCCEEDED (retcode))
-                stmt->bind_type = Attribute;
+                stmt->row_bind_type = (SQLUINTEGER) (SQLULEN) ValuePtr;
 	      return retcode;
 	    }
 	  else
@@ -1613,7 +1674,7 @@ SQLSetStmtAttr_Internal (
 	      CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
 		  (stmt->dhstmt, Attribute, ValuePtr));
               if (Attribute == SQL_ATTR_ROW_BIND_TYPE && SQL_SUCCEEDED (retcode))
-                stmt->bind_type = Attribute;
+                stmt->row_bind_type = (SQLUINTEGER) (SQLULEN) ValuePtr;
 	      return retcode;
 	    }
 	  else
@@ -1758,6 +1819,8 @@ SQLSetConnectAttr_Internal (
   SWORD unicode_driver = (penv ? penv->unicode_driver : 0);
   SQLUINTEGER odbc_ver;
   SQLUINTEGER dodbc_ver;
+  CONV_DIRECT conv_direct = CD_NONE;
+  DM_CONV *conv = &con->conv;
 
   odbc_ver = ((GENV_t *) con->genv)->odbc_ver;
   dodbc_ver = (penv != SQL_NULL_HENV) ? penv->dodbc_ver : odbc_ver;
@@ -1768,9 +1831,14 @@ SQLSetConnectAttr_Internal (
       return SQL_ERROR;
     }
 
-  if (penv && 
-      ((unicode_driver && waMode != 'W')
-       || (!unicode_driver && waMode == 'W')))
+  if (unicode_driver && waMode != 'W')
+    conv_direct = CD_A2W;
+  else if (!unicode_driver && waMode == 'W')
+    conv_direct = CD_W2A;
+  else if (waMode == 'W' && conv->dm_cp != conv->drv_cp)
+    conv_direct = CD_W2W;
+
+  if (conv_direct != CD_NONE)
     {
       switch(Attribute)
         {
@@ -1778,17 +1846,10 @@ SQLSetConnectAttr_Internal (
         case SQL_ATTR_TRACEFILE:
         case SQL_ATTR_TRANSLATE_LIB:
 
-          if (waMode != 'W')
-            {
-            /* ansi=>unicode*/
-              _ValuePtr = dm_SQL_A2W((SQLCHAR *) ValuePtr, StringLength);
-            }
-          else
-            {
-            /* unicode=>ansi*/
-              StringLength = (StringLength != SQL_NTS) ? (SQLINTEGER) (StringLength / sizeof(wchar_t)) : SQL_NTS;
-              _ValuePtr = dm_SQL_W2A((SQLWCHAR *) ValuePtr, StringLength);
-            }
+          if (waMode == 'W')
+            StringLength = (StringLength != SQL_NTS) ? (SQLINTEGER) (StringLength / DM_WCHARSIZE(conv)) : SQL_NTS;
+          _ValuePtr = conv_text_m2d (conv, ValuePtr, StringLength, conv_direct);
+
           ValuePtr = _ValuePtr;
           StringLength = SQL_NTS;
           break;
@@ -1801,10 +1862,29 @@ SQLSetConnectAttr_Internal (
       (  odbc_ver == SQL_OV_ODBC3 
        || (odbc_ver == SQL_OV_ODBC2 && hproc2 == SQL_NULL_HPROC)))
     {
-      CALL_UDRIVER(con, con, retcode, hproc, unicode_driver, 
+      CALL_UDRIVER(con, con, retcode, hproc, unicode_driver,
         en_SetConnectAttr, (con->dhdbc, Attribute, ValuePtr, StringLength));
       if (hproc != SQL_NULL_HPROC)
-        return retcode;
+        {
+          if (retcode == SQL_SUCCESS && Attribute == SQL_ATTR_APP_WCHAR_TYPE)
+            {
+              switch((SQLINTEGER)ValuePtr)
+                {
+                case CP_UTF16:
+                  penv->conv.drv_cp = (IODBC_CHARSET)ValuePtr;
+                  break;
+
+                case CP_UTF8:
+                  penv->conv.drv_cp = (IODBC_CHARSET)ValuePtr;
+                  break;
+
+                case CP_UCS4:
+                  penv->conv.drv_cp = (IODBC_CHARSET)ValuePtr;
+                  break;
+                }
+            }
+          return retcode;
+        }
     }
 
   switch (Attribute)
@@ -1922,6 +2002,9 @@ SQLGetConnectAttr_Internal (
   SWORD unicode_driver = (penv ? penv->unicode_driver : 0);
   SQLUINTEGER odbc_ver;
   SQLUINTEGER dodbc_ver;
+  CONV_DIRECT conv_direct = CD_NONE;
+  DM_CONV *conv = &con->conv;
+  SQLINTEGER _StringLength = StringLength;
 
   odbc_ver = ((GENV_t *) con->genv)->odbc_ver;
   dodbc_ver = (penv != SQL_NULL_HENV) ? penv->dodbc_ver : odbc_ver;
@@ -1932,37 +2015,45 @@ SQLGetConnectAttr_Internal (
       return SQL_ERROR;
     }
 
-  if (penv &&
-      ((unicode_driver && waMode != 'W') 
-       || (!unicode_driver && waMode == 'W')))
+  if (unicode_driver && waMode != 'W')
+    conv_direct = CD_A2W;
+  else if (!unicode_driver && waMode == 'W')
+    conv_direct = CD_W2A;
+  else if (waMode == 'W' && conv->dm_cp != conv->drv_cp)
+    conv_direct = CD_W2W;
+
+  if (penv && conv_direct != CD_NONE)
     {
       switch(Attribute)
         {
         case SQL_ATTR_CURRENT_CATALOG:
         case SQL_ATTR_TRACEFILE:
         case SQL_ATTR_TRANSLATE_LIB:
+          if (conv_direct == CD_A2W || conv_direct == CD_W2W)
+            {
+              /*ansi<=unicode*/
+              if (conv_direct == CD_W2W)
+                _StringLength /= DM_WCHARSIZE(conv);
 
-          if (waMode != 'W')
-            {
-            /* ansi=>unicode*/
-              StringLength *= sizeof(wchar_t);
-              if ((_Value = malloc(StringLength + 1)) == NULL)
+              if ((_Value = malloc((_StringLength + 1) * DRV_WCHARSIZE_ALLOC(conv))) == NULL)
 	        {
                   PUSHSQLERR (con->herr, en_HY001);
                   return SQL_ERROR;
                 }
+              _StringLength *= DRV_WCHARSIZE_ALLOC(conv);
+              valueOut = _Value;
             }
-          else
+          else if (conv_direct == CD_W2A)
             {
-            /* unicode=>ansi*/
-              StringLength /= sizeof(wchar_t);
-              if ((_Value = malloc(StringLength + 1)) == NULL)
+              /*unicode<=ansi*/
+              if ((_Value = malloc(StringLength * MB_CUR_MAX + 1)) == NULL)
 	        {
                   PUSHSQLERR (con->herr, en_HY001);
                   return SQL_ERROR;
                 }
+              _StringLength /= DM_WCHARSIZE(conv);
+              valueOut = _Value;
             }
-          valueOut = _Value;
           break;
         }
     }
@@ -1973,44 +2064,45 @@ SQLGetConnectAttr_Internal (
       (  odbc_ver == SQL_OV_ODBC3 
        || (odbc_ver == SQL_OV_ODBC2 && hproc2 == SQL_NULL_HPROC)))
     {
-      CALL_UDRIVER(con, con, retcode, hproc, unicode_driver, 
-        en_GetConnectAttr, (con->dhdbc, Attribute, valueOut, StringLength,
+      CALL_UDRIVER(con, con, retcode, hproc, unicode_driver,
+        en_GetConnectAttr, (con->dhdbc, Attribute, valueOut, _StringLength,
         StringLengthPtr));
     }
 
   if (hproc != SQL_NULL_HPROC)
     {
-      if (ValuePtr 
-          && SQL_SUCCEEDED (retcode)
-          && ((unicode_driver && waMode != 'W') 
-              || (!unicode_driver && waMode == 'W')))
+      if (ValuePtr && conv_direct != CD_NONE && SQL_SUCCEEDED (retcode))
         {
+          int count;
           switch(Attribute)
             {
             case SQL_ATTR_CURRENT_CATALOG:
             case SQL_ATTR_TRACEFILE:
             case SQL_ATTR_TRANSLATE_LIB:
-              if (waMode != 'W')
+              if (conv_direct == CD_A2W)
                 {
-                /* ansi<=unicode*/
-		  SQLSMALLINT retlen;
-
-                  dm_StrCopyOut2_W2A ((SQLWCHAR *) valueOut,
-			(SQLCHAR *) ValuePtr, 
-			StringLength / sizeof(wchar_t), &retlen);
+                  /*ansi<=unicode*/
+                  dm_StrCopyOut2_W2A_d2m (conv, valueOut,
+			(SQLCHAR *) ValuePtr,
+			StringLength, NULL, &count);
                   if (StringLengthPtr)
-                    *StringLengthPtr = retlen;
+                    *StringLengthPtr = (SQLSMALLINT)count;
                 }
-              else
+              else if (conv_direct == CD_W2A)
                 {
-                /* unicode<=ansi*/
-		  SQLSMALLINT retlen;
-
-                  dm_StrCopyOut2_A2W ((SQLCHAR *) valueOut, 
-			(SQLWCHAR *) ValuePtr, 
-			StringLength, &retlen);
+                  /*unicode<=ansi*/
+                  dm_StrCopyOut2_A2W_d2m (conv, (SQLCHAR *) valueOut,
+			ValuePtr, StringLength, NULL, &count);
                   if (StringLengthPtr)
-                    *StringLengthPtr = retlen * sizeof(wchar_t);
+                    *StringLengthPtr = (SQLSMALLINT)count;
+                }
+              else if (conv_direct == CD_W2W)
+                {
+                  /*unicode<=unicode*/
+                  dm_StrCopyOut2_W2W_d2m (conv, valueOut,
+			ValuePtr, StringLength, NULL, &count);
+                  if (StringLengthPtr)
+                    *StringLengthPtr = (SQLSMALLINT)count;
                 }
             }
         }
@@ -2035,7 +2127,7 @@ SQLGetConnectAttr_Internal (
 	    if (waMode != 'W')
 	      *StringLengthPtr = STRLEN (ValuePtr);
 	    else
-	      *StringLengthPtr = WCSLEN (ValuePtr) * sizeof(wchar_t);
+	      *StringLengthPtr = DM_WCSLEN (conv, ValuePtr) * DM_WCHARSIZE(conv);
 	  }
     }
   return retcode;
@@ -2142,9 +2234,18 @@ SQLGetDescField_Internal (
   SQLRETURN retcode = SQL_SUCCESS;
   void * valueOut = ValuePtr;
   void * _ValuePtr = NULL;
+  CONV_DIRECT conv_direct = CD_NONE;
+  DM_CONV *conv = &pdbc->conv;
+  SQLINTEGER _BufferLength = BufferLength;
 
-  if ((penv->unicode_driver && waMode != 'W') 
-      || (!penv->unicode_driver && waMode == 'W'))
+  if (penv->unicode_driver && waMode != 'W')
+    conv_direct = CD_A2W;
+  else if (!penv->unicode_driver && waMode == 'W')
+    conv_direct = CD_W2A;
+  else if (waMode == 'W' && conv->dm_cp != conv->drv_cp)
+    conv_direct = CD_W2W;
+
+  if (conv_direct != CD_NONE)
     {
       switch(FieldIdentifier)
         {
@@ -2159,47 +2260,49 @@ SQLGetDescField_Internal (
         case SQL_DESC_SCHEMA_NAME:
         case SQL_DESC_TABLE_NAME:
         case SQL_DESC_TYPE_NAME:
-          if (waMode != 'W')
+          if (conv_direct == CD_A2W || conv_direct == CD_W2W)
             {
-            /* ansi=>unicode*/
-              BufferLength *= sizeof(wchar_t);
-              if ((_ValuePtr = malloc(BufferLength + 1)) == NULL)
+             /*ansi<=unicode*/
+              if (conv_direct == CD_W2W)
+                _BufferLength /= DM_WCHARSIZE(conv);
+
+              if ((_ValuePtr = malloc((_BufferLength + 1) * DRV_WCHARSIZE_ALLOC(conv))) == NULL)
 	        {
                   PUSHSQLERR (desc->herr, en_HY001);
                   return SQL_ERROR;
                 }
+              _BufferLength *= DRV_WCHARSIZE_ALLOC(conv);
+              valueOut = _ValuePtr;
             }
-          else
+          else if (conv_direct == CD_W2A)
             {
-            /* unicode=>ansi*/
-              BufferLength /= sizeof(wchar_t);
-              if ((_ValuePtr = malloc(BufferLength + 1)) == NULL)
-    	        {
+              /*unicode<=ansi*/
+              if ((_ValuePtr = malloc(_BufferLength * MB_CUR_MAX + 1)) == NULL)
+	        {
                   PUSHSQLERR (desc->herr, en_HY001);
                   return SQL_ERROR;
                 }
+              _BufferLength /= DM_WCHARSIZE(conv);
+              valueOut = _ValuePtr;
             }
-          valueOut = _ValuePtr;
           break;
         }
     }
 
-  CALL_UDRIVER(desc->hdbc, desc, retcode, hproc, penv->unicode_driver, 
+  CALL_UDRIVER(desc->hdbc, desc, retcode, hproc, penv->unicode_driver,
     en_GetDescField, (desc->dhdesc, RecNumber, FieldIdentifier, valueOut,
-    BufferLength, StringLengthPtr));
+    _BufferLength, StringLengthPtr));
 
   if (hproc == SQL_NULL_HPROC)
     {
       MEM_FREE(_ValuePtr);
       PUSHSQLERR (desc->herr, en_IM001);
       return SQL_ERROR;
-    }    
+    }
 
-  if (ValuePtr 
-      && SQL_SUCCEEDED (retcode)
-      && ((penv->unicode_driver && waMode != 'W') 
-          || (!penv->unicode_driver && waMode == 'W')))
+  if (ValuePtr && conv_direct != CD_NONE && SQL_SUCCEEDED (retcode))
     {
+      int count;
       switch(FieldIdentifier)
         {
         case SQL_DESC_BASE_COLUMN_NAME:
@@ -2213,27 +2316,29 @@ SQLGetDescField_Internal (
         case SQL_DESC_SCHEMA_NAME:
         case SQL_DESC_TABLE_NAME:
         case SQL_DESC_TYPE_NAME:
-          if (waMode != 'W')
+          if (conv_direct == CD_A2W)
             {
             /* ansi<=unicode*/
-	      SQLSMALLINT retlen;
-
-              dm_StrCopyOut2_W2A ((SQLWCHAR *) valueOut, 
-		(SQLCHAR *) ValuePtr, 
-		BufferLength / sizeof(wchar_t), &retlen);
-	      if (StringLengthPtr)
-                *StringLengthPtr = retlen;
+              dm_StrCopyOut2_W2A_d2m (conv, valueOut,
+		(SQLCHAR *) ValuePtr, BufferLength, NULL, &count);
+              if (StringLengthPtr)
+                *StringLengthPtr = (SQLSMALLINT)count;
             }
-          else
+          else if (conv_direct == CD_W2A)
             {
             /* unicode<=ansi*/
-	      SQLSMALLINT retlen;
-
-              dm_StrCopyOut2_A2W ((SQLCHAR *) valueOut, 
-		(SQLWCHAR *) ValuePtr, 
-		BufferLength, &retlen);
+              dm_StrCopyOut2_A2W_d2m (conv, (SQLCHAR *) valueOut,
+		ValuePtr, BufferLength, NULL, &count);
               if (StringLengthPtr)
-                *StringLengthPtr = retlen * sizeof(wchar_t);
+                *StringLengthPtr = (SQLSMALLINT)count;
+            }
+          else if (conv_direct == CD_W2W)
+            {
+            /* unicode<=unicode*/
+              dm_StrCopyOut2_W2W_d2m (conv, valueOut,
+		ValuePtr, BufferLength, NULL, &count);
+              if (StringLengthPtr)
+                *StringLengthPtr = (SQLSMALLINT)count;
             }
           break;
         }
@@ -2356,9 +2461,17 @@ SQLSetDescField_Internal (
   HPROC hproc = SQL_NULL_HPROC;
   SQLRETURN retcode = SQL_SUCCESS;
   void * _ValuePtr = NULL;
+  CONV_DIRECT conv_direct = CD_NONE;
+  DM_CONV *conv = &pdbc->conv;
 
-  if ((penv->unicode_driver && waMode != 'W')
-      || (!penv->unicode_driver && waMode == 'W'))
+  if (penv->unicode_driver && waMode != 'W')
+    conv_direct = CD_A2W;
+  else if (!penv->unicode_driver && waMode == 'W')
+    conv_direct = CD_W2A;
+  else if (waMode == 'W' && conv->dm_cp != conv->drv_cp)
+    conv_direct = CD_W2W;
+
+  if (conv_direct != CD_NONE)
     {
       switch(FieldIdentifier)
         {
@@ -2373,17 +2486,9 @@ SQLSetDescField_Internal (
         case SQL_DESC_SCHEMA_NAME:
         case SQL_DESC_TABLE_NAME:
         case SQL_DESC_TYPE_NAME:
-          if (waMode != 'W')
-            {
-            /* ansi=>unicode*/
-              _ValuePtr = dm_SQL_A2W((SQLCHAR *) ValuePtr, BufferLength);
-            }
-          else
-            {
-            /* unicode=>ansi*/
-              BufferLength = (BufferLength != SQL_NTS) ? (SQLINTEGER) (BufferLength / sizeof(wchar_t)) : SQL_NTS;
-              _ValuePtr = dm_SQL_W2A((SQLWCHAR *) ValuePtr, BufferLength);
-            }
+          if (waMode == 'W')
+            BufferLength = (BufferLength != SQL_NTS) ? (SQLINTEGER) (BufferLength / DM_WCHARSIZE(conv)) : SQL_NTS;
+          _ValuePtr = conv_text_m2d (conv, ValuePtr, BufferLength, conv_direct);
           ValuePtr = _ValuePtr;
           BufferLength = SQL_NTS;
           break;
@@ -2498,26 +2603,33 @@ SQLGetDescRec_Internal (
   void * nameOut = Name;
   void * _Name = NULL;
 
-  if ((penv->unicode_driver && waMode != 'W') 
-      || (!penv->unicode_driver && waMode == 'W'))
+  CONV_DIRECT conv_direct = CD_NONE;
+  DM_CONV *conv = &pdbc->conv;
+
+  if (penv->unicode_driver && waMode != 'W')
+    conv_direct = CD_A2W;
+  else if (!penv->unicode_driver && waMode == 'W')
+    conv_direct = CD_W2A;
+  else if (waMode == 'W' && conv->dm_cp != conv->drv_cp)
+    conv_direct = CD_W2W;
+
+  if (conv_direct == CD_A2W || conv_direct == CD_W2W)
     {
-      if (waMode != 'W')
+      /* ansi<=unicode*/
+      if ((_Name = malloc((BufferLength + 1) * DRV_WCHARSIZE_ALLOC(conv))) == NULL)
         {
-        /* ansi=>unicode*/
-          if ((_Name = malloc(BufferLength * sizeof(wchar_t) + 1)) == NULL)
-            {
-              PUSHSQLERR (desc->herr, en_HY001);
-              return SQL_ERROR;
-            }
+          PUSHSQLERR (desc->herr, en_HY001);
+          return SQL_ERROR;
         }
-      else
+      nameOut = _Name;
+    }
+  else if (conv_direct == CD_W2A)
+    {
+      /* unicode<=ansi*/
+      if ((_Name = malloc(BufferLength * MB_CUR_MAX + 1)) == NULL)
         {
-        /* unicode=>ansi*/
-          if ((_Name = malloc(BufferLength + 1)) == NULL)
-            {
-              PUSHSQLERR (desc->herr, en_HY001);
-              return SQL_ERROR;
-            }
+          PUSHSQLERR (desc->herr, en_HY001);
+          return SQL_ERROR;
         }
       nameOut = _Name;
     }
@@ -2532,22 +2644,34 @@ SQLGetDescRec_Internal (
       MEM_FREE(_Name);
       PUSHSQLERR (desc->herr, en_IM001);
       return SQL_ERROR;
-    }    
+    }
 
-  if (Name 
-      && SQL_SUCCEEDED (retcode)
-      && ((penv->unicode_driver && waMode != 'W') 
-          || (!penv->unicode_driver && waMode == 'W')))
+  if (Name && conv_direct != CD_NONE && SQL_SUCCEEDED (retcode))
     {
-      if (waMode != 'W')
+      int count;
+      if (conv_direct == CD_A2W)
         {
         /* ansi<=unicode*/
-          dm_StrCopyOut2_W2A ((SQLWCHAR *) nameOut, (SQLCHAR *) Name, BufferLength, StringLengthPtr);
+          dm_StrCopyOut2_W2A_d2m (conv, nameOut, (SQLCHAR *) Name,
+              BufferLength, NULL, &count);
+          if (StringLengthPtr)
+            *StringLengthPtr = (SQLSMALLINT)count;
         }
-      else
+      else if (conv_direct == CD_W2A)
         {
         /* unicode<=ansi*/
-          dm_StrCopyOut2_A2W ((SQLCHAR *) nameOut, (SQLWCHAR *) Name, BufferLength, StringLengthPtr);
+          dm_StrCopyOut2_A2W_d2m (conv, (SQLCHAR *) nameOut, Name,
+              BufferLength, NULL, &count);
+          if (StringLengthPtr)
+            *StringLengthPtr = (SQLSMALLINT)count;
+        }
+      else if (conv_direct == CD_W2W)
+        {
+        /* unicode<=unicode*/
+          dm_StrCopyOut2_W2W_d2m (conv, nameOut, Name, BufferLength,
+              NULL, &count);
+          if (StringLengthPtr)
+            *StringLengthPtr = (SQLSMALLINT)count;
         }
     }
 
@@ -2840,12 +2964,21 @@ SQLColAttribute_Internal (
   void * _charAttr = NULL;
   SQLUINTEGER odbc_ver;
   SQLUINTEGER dodbc_ver;
+  CONV_DIRECT conv_direct = CD_NONE;
+  DM_CONV *conv = &pdbc->conv;
+  SQLSMALLINT _BufferLength = BufferLength;
 
   odbc_ver = ((GENV_t *) pdbc->genv)->odbc_ver;
   dodbc_ver = (penv != SQL_NULL_HENV) ? penv->dodbc_ver : odbc_ver;
 
-  if ((penv->unicode_driver && waMode != 'W') 
-      || (!penv->unicode_driver && waMode == 'W'))
+  if (penv->unicode_driver && waMode != 'W')
+    conv_direct = CD_A2W;
+  else if (!penv->unicode_driver && waMode == 'W')
+    conv_direct = CD_W2A;
+  else if (waMode == 'W' && conv->dm_cp != conv->drv_cp)
+    conv_direct = CD_W2W;
+
+  if (conv_direct != CD_NONE)
     {
       switch(FieldIdentifier)
         {
@@ -2862,27 +2995,31 @@ SQLColAttribute_Internal (
         case SQL_DESC_SCHEMA_NAME:
         case SQL_DESC_TABLE_NAME:
         case SQL_DESC_TYPE_NAME:
-          if (waMode != 'W')
+          if (conv_direct == CD_A2W || conv_direct == CD_W2W)
             {
-            /* ansi=>unicode*/
-              BufferLength *= sizeof(wchar_t);
-              if ((_charAttr = malloc(BufferLength + 1)) == NULL)
+              /*ansi<=unicode*/
+              if (conv_direct == CD_W2W)
+                _BufferLength /= DM_WCHARSIZE(conv);
+
+              if ((_charAttr = malloc((_BufferLength + 1) * DRV_WCHARSIZE_ALLOC(conv))) == NULL)
                 {
                   PUSHSQLERR (stmt->herr, en_HY001);
                   return SQL_ERROR;
                 }
+              _BufferLength *= DRV_WCHARSIZE_ALLOC(conv);
+              charAttrOut = _charAttr;
             }
-          else
+          else if (conv_direct == CD_W2A)
             {
-            /* unicode=>ansi*/
-              BufferLength /= sizeof(wchar_t);
-              if ((_charAttr = malloc(BufferLength + 1)) == NULL)
+              /*unicode<=ansi*/
+              if ((_charAttr = malloc(BufferLength * MB_CUR_MAX + 1)) == NULL)
                 {
                   PUSHSQLERR (stmt->herr, en_HY001);
                   return SQL_ERROR;
                 }
+              _BufferLength /= DM_WCHARSIZE(conv);
+              charAttrOut = _charAttr;
             }
-          charAttrOut = _charAttr;
           break;
         }
     }
@@ -2893,9 +3030,9 @@ SQLColAttribute_Internal (
       (  odbc_ver == SQL_OV_ODBC3 
        || (odbc_ver == SQL_OV_ODBC2 && hproc2 == SQL_NULL_HPROC)))
     {
-      CALL_UDRIVER(stmt->hdbc, stmt, retcode, hproc, penv->unicode_driver, 
-        en_ColAttribute, (stmt->dhstmt, ColumnNumber, FieldIdentifier, 
-        charAttrOut, BufferLength, StringLengthPtr, NumericAttributePtr));
+      CALL_UDRIVER(stmt->hdbc, stmt, retcode, hproc, penv->unicode_driver,
+        en_ColAttribute, (stmt->dhstmt, ColumnNumber, FieldIdentifier,
+        charAttrOut, _BufferLength, StringLengthPtr, NumericAttributePtr));
     }
 
   if (hproc != SQL_NULL_HPROC)
@@ -2911,11 +3048,10 @@ SQLColAttribute_Internal (
 	  *ptr = _iodbcdm_map_sql_type (*ptr, genv->odbc_ver);
       }
 
-      if (CharacterAttributePtr 
-          && SQL_SUCCEEDED (retcode)
-          &&  ((penv->unicode_driver && waMode != 'W') 
-              || (!penv->unicode_driver && waMode == 'W')))
+      if (CharacterAttributePtr && conv_direct != CD_NONE
+          && SQL_SUCCEEDED (retcode))
         {
+          int count;
           switch(FieldIdentifier)
             {
             case SQL_COLUMN_NAME:
@@ -2930,21 +3066,30 @@ SQLColAttribute_Internal (
             case SQL_DESC_SCHEMA_NAME:
             case SQL_DESC_TABLE_NAME:
             case SQL_DESC_TYPE_NAME:
-              if (waMode != 'W')
+              if (conv_direct == CD_A2W)
                 {
-                /* ansi<=unicode*/
-                  dm_StrCopyOut2_W2A ((SQLWCHAR *) charAttrOut, 
+                  /*ansi<=unicode*/
+                  dm_StrCopyOut2_W2A_d2m (conv, charAttrOut,
 			(SQLCHAR *) CharacterAttributePtr,
-			BufferLength / sizeof(wchar_t), StringLengthPtr);
-                }
-              else
-                {
-                /* unicode<=ansi*/
-                  dm_StrCopyOut2_A2W ((SQLCHAR *) charAttrOut, 
-			(SQLWCHAR *) CharacterAttributePtr,
-			BufferLength, StringLengthPtr);
+			BufferLength, NULL, &count);
                   if (StringLengthPtr)
-                    *StringLengthPtr = *StringLengthPtr * sizeof(wchar_t);
+                    *StringLengthPtr = (SQLSMALLINT)count;
+                }
+              else if (conv_direct == CD_W2A)
+                {
+                  /*unicode<=ansi*/
+                  dm_StrCopyOut2_A2W_d2m (conv, (SQLCHAR *) charAttrOut,
+			CharacterAttributePtr, BufferLength, NULL, &count);
+                  if (StringLengthPtr)
+                    *StringLengthPtr = (SQLSMALLINT)count;
+                }
+              else if (conv_direct == CD_W2W)
+                {
+                  /*unicode<=unicode*/
+                  dm_StrCopyOut2_W2W_d2m (conv, charAttrOut,
+			CharacterAttributePtr, BufferLength, NULL, &count);
+                  if (StringLengthPtr)
+                    *StringLengthPtr = (SQLSMALLINT)count;
                 }
             }
         }
@@ -3097,9 +3242,9 @@ SQLColAttribute_Internal (
 	  return SQL_ERROR;
 	}
 
-      CALL_UDRIVER(stmt->hdbc, stmt, retcode, hproc, penv->unicode_driver, 
-        en_ColAttributes, (stmt->dhstmt, ColumnNumber, FieldIdentifier, 
-        charAttrOut, BufferLength, StringLengthPtr, NumericAttributePtr));
+      CALL_UDRIVER(stmt->hdbc, stmt, retcode, hproc, penv->unicode_driver,
+        en_ColAttributes, (stmt->dhstmt, ColumnNumber, FieldIdentifier,
+        charAttrOut, _BufferLength, StringLengthPtr, NumericAttributePtr));
 
       if (hproc == SQL_NULL_HPROC)
         {
@@ -3119,11 +3264,10 @@ SQLColAttribute_Internal (
 	  *ptr = _iodbcdm_map_sql_type (*ptr, genv->odbc_ver);
       }
 
-      if (CharacterAttributePtr 
-          && SQL_SUCCEEDED (retcode)
-          && ((penv->unicode_driver && waMode != 'W') 
-              || (!penv->unicode_driver && waMode == 'W')))
+      if (CharacterAttributePtr && conv_direct != CD_NONE
+          && SQL_SUCCEEDED (retcode))
         {
+          int count;
           switch(FieldIdentifier)
             {
             case SQL_COLUMN_QUALIFIER_NAME:
@@ -3139,21 +3283,30 @@ SQLColAttribute_Internal (
             case SQL_DESC_LITERAL_SUFFIX:
             case SQL_DESC_LOCAL_TYPE_NAME:
             case SQL_DESC_NAME:
-              if (waMode != 'W')
+              if (conv_direct == CD_A2W)
                 {
-                /* ansi<=unicode*/
-                  dm_StrCopyOut2_W2A ((SQLWCHAR *) charAttrOut, 
+                  /*ansi<=unicode*/
+                  dm_StrCopyOut2_W2A_d2m (conv, charAttrOut,
 			(SQLCHAR *) CharacterAttributePtr,
-			BufferLength / sizeof(wchar_t), StringLengthPtr);
-                }
-              else
-                {
-                /* unicode<=ansi*/
-                  dm_StrCopyOut2_A2W ((SQLCHAR *) charAttrOut,
-			(SQLWCHAR *) CharacterAttributePtr,
-			BufferLength, StringLengthPtr);
+			BufferLength, NULL, &count);
                   if (StringLengthPtr)
-                    *StringLengthPtr = *StringLengthPtr * sizeof(wchar_t);
+                    *StringLengthPtr = (SQLSMALLINT)count;
+                }
+              else if (conv_direct == CD_W2A)
+                {
+                  /*unicode<=ansi*/
+                  dm_StrCopyOut2_A2W_d2m (conv, (SQLCHAR *) charAttrOut,
+			CharacterAttributePtr, BufferLength, NULL, &count);
+                  if (StringLengthPtr)
+                    *StringLengthPtr = (SQLSMALLINT)count;
+                }
+              else if (conv_direct == CD_W2W)
+                {
+                  /*unicode<=unicode*/
+                  dm_StrCopyOut2_W2W_d2m (conv, charAttrOut,
+			CharacterAttributePtr, BufferLength, NULL, &count);
+                  if (StringLengthPtr)
+                    *StringLengthPtr = (SQLSMALLINT)count;
                 }
             }
         }
@@ -3332,14 +3485,25 @@ SQLBulkOperations_Internal (
       return SQL_ERROR;
     }
 
+  if (Operation == SQL_FETCH_BY_BOOKMARK )
+    {
+      if ((retcode = _iodbcdm_FixColBindData (stmt)) != SQL_SUCCESS)
+        return retcode;
+    }
+
+  if (Operation == SQL_ADD || Operation == SQL_UPDATE_BY_BOOKMARK
+      || Operation == SQL_DELETE_BY_BOOKMARK)
+    {
+      _iodbcdm_ConvBindData_m2d (stmt);
+    }
+
   hproc = _iodbcdm_getproc (stmt->hdbc, en_BulkOperations);
   if (hproc)
     {
       CALL_DRIVER (stmt->hdbc, stmt, retcode, hproc,
 	  (stmt->dhstmt, Operation));
-      
-      if (Operation == SQL_FETCH_BY_BOOKMARK 
-          && SQL_SUCCEEDED (retcode))
+
+      if (Operation == SQL_FETCH_BY_BOOKMARK && SQL_SUCCEEDED (retcode))
         _iodbcdm_ConvBindData (stmt);
       return retcode;
     }
@@ -3411,7 +3575,6 @@ SQLFetchScroll_Internal (
 	{
 	case en_stmt_allocated:
 	case en_stmt_prepared:
-	case en_stmt_fetched:
 	case en_stmt_needdata:
 	case en_stmt_mustput:
 	case en_stmt_canput:
@@ -3490,6 +3653,7 @@ SQLFetchScroll_Internal (
   switch (stmt->state)
     {
     case en_stmt_cursoropen:
+    case en_stmt_fetched:
     case en_stmt_xfetched:
       switch (retcode)
 	{
@@ -3526,12 +3690,16 @@ SQLFetchScroll (
   ENTER_STMT (StatementHandle,
     trace_SQLFetchScroll (TRACE_ENTER,
     	StatementHandle, 
-	FetchOrientation, 
+	FetchOrientation,
 	FetchOffset));
 
-  retcode = SQLFetchScroll_Internal ( 
-  	StatementHandle, 
-	FetchOrientation, 
+  retcode = _iodbcdm_FixColBindData ((STMT_t *) StatementHandle);
+  if (retcode != SQL_SUCCESS)
+    return retcode;
+
+  retcode = SQLFetchScroll_Internal (
+	StatementHandle,
+	FetchOrientation,
 	FetchOffset);
 
   if (SQL_SUCCEEDED (retcode))
