@@ -8,7 +8,7 @@
  *  The iODBC driver manager.
  *
  *  Copyright (C) 1995 Ke Jin <kejin@empress.com>
- *  Copyright (C) 1996-2023 OpenLink Software <iodbc@openlinksw.com>
+ *  Copyright (C) 1996-2021 OpenLink Software <iodbc@openlinksw.com>
  *  All Rights Reserved.
  *
  *  This software is released under the terms of either of the following
@@ -82,16 +82,16 @@
 #include <sqlext.h>
 #include <sqlucode.h>
 
-#include "unicode.h"
+#include <unicode.h>
 
-#include "dlproc.h"
+#include <dlproc.h>
 
-#include "herr.h"
-#include "henv.h"
-#include "hdbc.h"
-#include "hstmt.h"
+#include <herr.h>
+#include <henv.h>
+#include <hdbc.h>
+#include <hstmt.h>
 
-#include "itrace.h"
+#include <itrace.h>
 
 SQLRETURN SQL_API
 SQLPrepare_Internal (
