@@ -7,8 +7,8 @@
  *
  *  The iODBC driver manager.
  *
- *  Copyright (C) 1995 by Ke Jin <kejin@empress.com>
- *  Copyright (C) 1996-2016 by OpenLink Software <iodbc@openlinksw.com>
+ *  Copyright (C) 1995 Ke Jin <kejin@empress.com>
+ *  Copyright (C) 1996-2021 OpenLink Software <iodbc@openlinksw.com>
  *  All Rights Reserved.
  *
  *  This software is released under the terms of either of the following
@@ -198,8 +198,7 @@ SQLBindCol_Internal (
     CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc,
         (pstmt->dhstmt, icol, nCType, rgbValue, cbValueMax, pcbValue));
 
-  if (icol != 0 && !penv->unicode_driver && nCType == SQL_C_WCHAR 
-      && SQL_SUCCEEDED (retcode))
+  if (SQL_SUCCEEDED (retcode))
     {
       BIND_t tbind;
 
@@ -208,7 +207,17 @@ SQLBindCol_Internal (
       tbind.bn_data = rgbValue;
       tbind.bn_size = cbValueMax;
       tbind.bn_pInd = pcbValue;
-      
+      tbind.direct  = CD_NONE;
+      tbind.bn_conv_size = 0;
+      tbind.bn_conv_data = NULL;
+      tbind.bn_conv_pInd = NULL;
+      tbind.bn_tmp = NULL;
+      tbind.bn_tmp_Ind = NULL;
+      tbind.rebinded = FALSE;
+
+      if (nCType == SQL_C_WCHAR)
+        tbind.direct  = penv->unicode_driver ? CD_W2W : CD_A2W;
+
       if (rgbValue)
         _iodbcdm_BindColumn (pstmt, &tbind);
       else
@@ -256,6 +265,8 @@ SQLGetCursorName_Internal (
   SQLRETURN retcode = SQL_SUCCESS;
   void * cursorOut = szCursor;
   void * _Cursor = NULL;
+  CONV_DIRECT conv_direct = CD_NONE;
+  DM_CONV *conv = &pdbc->conv;
 
   /* check argument */
   if (cbCursorMax < (SWORD) 0)
@@ -281,29 +292,31 @@ SQLGetCursorName_Internal (
       return SQL_ERROR;
     }
 
+  if (penv->unicode_driver && waMode != 'W')
+    conv_direct = CD_A2W;
+  else if (!penv->unicode_driver && waMode == 'W')
+    conv_direct = CD_W2A;
+  else if (waMode == 'W' && conv->dm_cp != conv->drv_cp)
+    conv_direct = CD_W2W;
+
   /* call driver's function */
-
-
-  if ((penv->unicode_driver && waMode != 'W') 
-      || (!penv->unicode_driver && waMode == 'W'))
+  if (conv_direct == CD_A2W || conv_direct == CD_W2W)
     {
-      if (waMode != 'W')
-        {
-        /* ansi=>unicode*/
-          if ((_Cursor = malloc((cbCursorMax + 1) * sizeof(wchar_t))) == NULL)
-	    {
-              PUSHSQLERR (pstmt->herr, en_HY001);
-              return SQL_ERROR;
-            }
+      /* ansi<=unicode*/
+      if ((_Cursor = malloc((cbCursorMax + 1) * DRV_WCHARSIZE_ALLOC(conv))) == NULL)
+	{
+          PUSHSQLERR (pstmt->herr, en_HY001);
+          return SQL_ERROR;
         }
-      else
-        {
-        /* unicode=>ansi*/
-          if ((_Cursor = malloc(cbCursorMax + 1)) == NULL)
-	    {
-              PUSHSQLERR (pstmt->herr, en_HY001);
-              return SQL_ERROR;
-            }
+      cursorOut = _Cursor;
+    }
+  else if (conv_direct == CD_W2A)
+    {
+      /* unicode<=ansi*/
+      if ((_Cursor = malloc(cbCursorMax * MB_CUR_MAX + 1)) == NULL)
+	{
+          PUSHSQLERR (pstmt->herr, en_HY001);
+          return SQL_ERROR;
         }
       cursorOut = _Cursor;
     }
@@ -323,20 +336,26 @@ SQLGetCursorName_Internal (
       return SQL_ERROR;
     }
 
-  if (szCursor 
-      && SQL_SUCCEEDED (retcode)
-      && ((penv->unicode_driver && waMode != 'W') 
-          || (!penv->unicode_driver && waMode == 'W')))
+  if (szCursor && conv_direct != CD_NONE
+      && SQL_SUCCEEDED (retcode))
     {
-      if (waMode != 'W')
+      if (conv_direct == CD_A2W)
         {
-        /* ansi<=unicode*/
-          dm_StrCopyOut2_W2A ((SQLWCHAR *) cursorOut, (SQLCHAR *) szCursor, cbCursorMax, NULL);
+         /*ansi<=unicode*/
+          dm_StrCopyOut2_W2A_d2m (conv, cursorOut, (SQLCHAR *) szCursor,
+		cbCursorMax, NULL, NULL);
         }
-      else
+      else if (conv_direct == CD_W2A)
         {
-        /* unicode<=ansi*/
-          dm_StrCopyOut2_A2W ((SQLCHAR *) cursorOut, (SQLWCHAR *) szCursor, cbCursorMax, NULL);
+         /*unicode<=ansi*/
+          dm_StrCopyOut2_A2W_d2m (conv, (SQLCHAR *) cursorOut, szCursor,
+		cbCursorMax * DM_WCHARSIZE(conv), NULL, NULL);
+        }
+      else if (conv_direct == CD_W2W)
+        {
+         /*unicode<=unicode*/
+          dm_StrCopyOut2_W2W_d2m (conv, cursorOut, szCursor,
+		cbCursorMax * DM_WCHARSIZE(conv), NULL, NULL);
         }
     }
 
@@ -632,6 +651,8 @@ SQLDescribeCol_Internal (
   void * _ColName = NULL;
   void * colNameOut = szColName;
   sqlstcode_t sqlstat = en_00000;
+  CONV_DIRECT conv_direct = CD_NONE;
+  DM_CONV *conv = &pdbc->conv;
 
   /* check arguments */
   if (icol == 0)
@@ -672,30 +693,33 @@ SQLDescribeCol_Internal (
       return SQL_ERROR;
     }
 
+  if (penv->unicode_driver && waMode != 'W')
+    conv_direct = CD_A2W;
+  else if (!penv->unicode_driver && waMode == 'W')
+    conv_direct = CD_W2A;
+  else if (waMode == 'W' && conv->dm_cp != conv->drv_cp)
+    conv_direct = CD_W2W;
+
   /* call driver */
-    if (szColName != NULL && cbColNameMax > 0 && 
-    	((penv->unicode_driver && waMode != 'W') || 
-	(!penv->unicode_driver && waMode == 'W'))
-	)
+
+  if (conv_direct == CD_A2W || conv_direct == CD_W2W)
     {
-      if (waMode != 'W')
-        {
-        /* ansi=>unicode*/
-          if ((_ColName = _iodbcdm_alloc_var(pstmt, 0, 
-          	             cbColNameMax * sizeof(wchar_t))) == NULL)
-	    {
-              PUSHSQLERR (pstmt->herr, en_HY001);
-              return SQL_ERROR;
-            }
+      /*ansi<=unicode*/
+      if ((_ColName = _iodbcdm_alloc_var(pstmt, 0,
+                     cbColNameMax * DRV_WCHARSIZE_ALLOC(conv))) == NULL)
+	{
+          PUSHSQLERR (pstmt->herr, en_HY001);
+          return SQL_ERROR;
         }
-      else
-        {
-        /* unicode=>ansi*/
-          if ((_ColName = _iodbcdm_alloc_var(pstmt, 0, cbColNameMax)) == NULL)
-	    {
-              PUSHSQLERR (pstmt->herr, en_HY001);
-              return SQL_ERROR;
-            }
+      colNameOut = _ColName;
+    }
+  else if (conv_direct == CD_W2A)
+    {
+      /*unicode<=ansi*/
+      if ((_ColName = _iodbcdm_alloc_var(pstmt, 0, cbColNameMax * MB_CUR_MAX + 1)) == NULL)
+	{
+          PUSHSQLERR (pstmt->herr, en_HY001);
+          return SQL_ERROR;
         }
       colNameOut = _ColName;
     }
@@ -726,20 +750,26 @@ SQLDescribeCol_Internal (
   if (SQL_SUCCEEDED(retcode) && pfSqlType)
     *pfSqlType = _iodbcdm_map_sql_type (*pfSqlType, genv->odbc_ver);
 
-  if (szColName 
-      && SQL_SUCCEEDED (retcode)
-      && ((penv->unicode_driver && waMode != 'W') 
-          || (!penv->unicode_driver && waMode == 'W')))
+  if (szColName && conv_direct != CD_NONE
+      && SQL_SUCCEEDED (retcode))
     {
-      if (waMode != 'W')
+      if (conv_direct == CD_A2W)
         {
-        /* ansi<=unicode*/
-          dm_StrCopyOut2_W2A ((SQLWCHAR *) colNameOut, (SQLCHAR *) szColName, cbColNameMax, NULL);
+         /*ansi<=unicode*/
+          dm_StrCopyOut2_W2A_d2m (conv, colNameOut, (SQLCHAR *) szColName,
+		cbColNameMax, NULL, NULL);
         }
-      else
+      else if (conv_direct == CD_W2A)
         {
-        /* unicode<=ansi*/
-          dm_StrCopyOut2_A2W ((SQLCHAR *) colNameOut, (SQLWCHAR *) szColName, cbColNameMax, NULL);
+         /*unicode<=ansi*/
+          dm_StrCopyOut2_A2W_d2m (conv, (SQLCHAR *) colNameOut, szColName,
+		cbColNameMax * DM_WCHARSIZE(conv), NULL, NULL);
+        }
+      else if (conv_direct == CD_W2W)
+        {
+         /*unicode<=unicode*/
+          dm_StrCopyOut2_W2W_d2m (conv, colNameOut, szColName,
+		cbColNameMax * DM_WCHARSIZE(conv), NULL, NULL);
         }
     }
 
@@ -937,6 +967,9 @@ SQLColAttributes_Internal (
   SQLUSMALLINT new_attr = fDescType;
   SQLUINTEGER odbc_ver = ((GENV_t *) pdbc->genv)->odbc_ver;
   SQLUINTEGER dodbc_ver = ((ENV_t *) pdbc->henv)->dodbc_ver;
+  CONV_DIRECT conv_direct = CD_NONE;
+  DM_CONV *conv = &pdbc->conv;
+  SQLSMALLINT _cbDescMax = cbDescMax;
 
   /* check arguments */
   if (icol == 0 && fDescType != SQL_COLUMN_COUNT)
@@ -984,10 +1017,15 @@ SQLColAttributes_Internal (
       return SQL_ERROR;
     }
 
+  if (penv->unicode_driver && waMode != 'W')
+    conv_direct = CD_A2W;
+  else if (!penv->unicode_driver && waMode == 'W')
+    conv_direct = CD_W2A;
+  else if (waMode == 'W' && conv->dm_cp != conv->drv_cp)
+    conv_direct = CD_W2W;
+
   /* call driver */
-  if (rgbDesc != NULL && cbDescMax > 0 && 
-      ((penv->unicode_driver && waMode != 'W') || 
-       (!penv->unicode_driver && waMode == 'W')))
+  if (conv_direct != CD_NONE)
     {
       switch(fDescType)
         {
@@ -998,27 +1036,32 @@ SQLColAttributes_Internal (
         case SQL_COLUMN_TABLE_NAME:
         case SQL_COLUMN_TYPE_NAME:
 
-          if (waMode != 'W')
+          if (conv_direct == CD_A2W || conv_direct == CD_W2W)
             {
-            /* ansi=>unicode*/
-              cbDescMax *= sizeof(wchar_t);
-              if ((_Desc = _iodbcdm_alloc_var(pstmt, 0, cbDescMax)) == NULL)
+             /*ansi<=unicode*/
+              if (conv_direct == CD_W2W)
+                _cbDescMax /= DM_WCHARSIZE(conv);
+
+              if ((_Desc = _iodbcdm_alloc_var(pstmt, 0,
+		        (_cbDescMax + 1) * DRV_WCHARSIZE_ALLOC(conv))) == NULL)
 	        {
                   PUSHSQLERR (pstmt->herr, en_HY001);
                   return SQL_ERROR;
                 }
+              _cbDescMax *= DRV_WCHARSIZE_ALLOC(conv);
+              descOut = _Desc;
             }
-          else
+          else if (conv_direct == CD_W2A)
             {
-            /* unicode=>ansi*/
-              cbDescMax /= sizeof(wchar_t);
-              if ((_Desc = _iodbcdm_alloc_var(pstmt, 0, cbDescMax)) == NULL)
+             /*unicode<=ansi*/
+              if ((_Desc = _iodbcdm_alloc_var(pstmt, 0, _cbDescMax * MB_CUR_MAX + 1)) == NULL)
 	        {
                   PUSHSQLERR (pstmt->herr, en_HY001);
                   return SQL_ERROR;
                 }
+              _cbDescMax /= DM_WCHARSIZE(conv);
+              descOut = _Desc;
             }
-          descOut = _Desc;
           break;
         }
     }
@@ -1068,12 +1111,12 @@ SQLColAttributes_Internal (
 
           CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc2, (
       	          pstmt->dhstmt, 
-      	          icol, 
-      	          fDescType, 
-      	          descOut, 
-      	          cbDescMax, 
-      	          pcbDesc, 
-      	          pfDesc));
+	          icol,
+	          fDescType,
+	          descOut,
+	          _cbDescMax,
+	          pcbDesc,
+	          pfDesc));
         }
     }
   else
@@ -1099,12 +1142,12 @@ SQLColAttributes_Internal (
         {
           CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc3, (
        	          pstmt->dhstmt, 
-       	          icol, 
-       	          new_attr, 
-       	          descOut, 
-       	          cbDescMax, 
-       	          pcbDesc, 
-       	          pfDesc));
+	          icol,
+	          new_attr,
+	          descOut,
+	          _cbDescMax,
+	          pcbDesc,
+	          pfDesc));
         }
       else
 #endif
@@ -1118,20 +1161,18 @@ SQLColAttributes_Internal (
 
           CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc2, (
       	          pstmt->dhstmt, 
-      	          icol, 
-      	          fDescType, 
-      	          descOut, 
-      	          cbDescMax, 
-      	          pcbDesc, 
-      	          pfDesc));
+	          icol,
+	          fDescType,
+	          descOut,
+	          _cbDescMax,
+	          pcbDesc,
+	          pfDesc));
         }
     }
 
-  if (rgbDesc 
-      && SQL_SUCCEEDED (retcode)
-      && ((penv->unicode_driver && waMode != 'W') 
-          || (!penv->unicode_driver && waMode == 'W')))
+  if (rgbDesc && conv_direct != CD_NONE && SQL_SUCCEEDED (retcode))
     {
+      int count;
       switch(fDescType)
         {
         case SQL_COLUMN_QUALIFIER_NAME:
@@ -1140,17 +1181,29 @@ SQLColAttributes_Internal (
         case SQL_COLUMN_OWNER_NAME:
         case SQL_COLUMN_TABLE_NAME:
         case SQL_COLUMN_TYPE_NAME:
-          if (waMode != 'W')
+          if (conv_direct == CD_A2W)
             {
-            /* ansi<=unicode*/
-              dm_StrCopyOut2_W2A ((SQLWCHAR *) descOut, (SQLCHAR *) rgbDesc, cbDescMax / sizeof(wchar_t), pcbDesc);
-            }
-          else
-            {
-            /* unicode<=ansi*/
-              dm_StrCopyOut2_A2W ((SQLCHAR *) descOut, (SQLWCHAR *) rgbDesc, cbDescMax, pcbDesc);
+             /*ansi<=unicode*/
+              dm_StrCopyOut2_W2A_d2m (conv, descOut, (SQLCHAR *) rgbDesc,
+		cbDescMax, NULL, &count);
               if (pcbDesc)
-                *pcbDesc = *pcbDesc * sizeof(wchar_t);
+                *pcbDesc = (SQLSMALLINT)count;
+            }
+          else if (conv_direct == CD_W2A)
+            {
+             /*unicode<=ansi*/
+              dm_StrCopyOut2_A2W_d2m (conv, (SQLCHAR *) descOut, rgbDesc,
+		cbDescMax, NULL, &count);
+              if (pcbDesc)
+                *pcbDesc = (SQLSMALLINT)count;
+            }
+          else if (conv_direct == CD_W2W)
+            {
+             /*unicode<=unicode*/
+              dm_StrCopyOut2_W2W_d2m (conv, descOut, rgbDesc,
+                cbDescMax, NULL, &count);
+              if (pcbDesc)
+                *pcbDesc = (SQLSMALLINT)count;
             }
         }
     }
