@@ -7,7 +7,7 @@
  *
  *  The iODBC driver manager.
  *
- *  Copyright (C) 1996-2023 OpenLink Software <iodbc@openlinksw.com>
+ *  Copyright (C) 1996-2021 OpenLink Software <iodbc@openlinksw.com>
  *  All Rights Reserved.
  *
  *  This software is released under the terms of either of the following
@@ -77,9 +77,9 @@
 
 #include "gui.h"
 
-#include "herr.h"
-#include "unicode.h"
-#include "dlproc.h"
+#include <herr.h>
+#include <unicode.h>
+#include <dlproc.h>
 
 #ifndef WIN32
 #include <unistd.h>
